@@ -3,6 +3,7 @@ from mujoco_py import MjSim, MjRenderContextOffscreen
 from mujoco_py import load_model_from_xml
 
 from robosuite.utils import SimulationError, XMLError, MujocoPyRenderer
+import robosuite.utils.macros as macros
 from robosuite.models.base import MujocoModel
 
 import numpy as np
@@ -120,6 +121,8 @@ class MujocoEnv(metaclass=EnvMeta):
         self.viewer = None
 
         # Simulation-specific attributes
+        self._observables = {}                      # Maps observable names to observable objects
+        self._obs_cache = {}                        # Maps observable names to pre-/partially-computed observable values
         self.control_freq = control_freq
         self.horizon = horizon
         self.ignore_done = ignore_done
@@ -143,6 +146,9 @@ class MujocoEnv(metaclass=EnvMeta):
         # Run all further internal (re-)initialization required
         self._reset_internal()
 
+        # Load observables
+        self._observables = self._setup_observables()
+
     def initialize_time(self, control_freq):
         """
         Initializes the time constants used for simulation.
@@ -151,14 +157,12 @@ class MujocoEnv(metaclass=EnvMeta):
             control_freq (float): Hz rate to run control loop at within the simulation
         """
         self.cur_time = 0
-        self.model_timestep = self.sim.model.opt.timestep
+        self.model_timestep = macros.SIMULATION_TIMESTEP
         if self.model_timestep <= 0:
-            raise XMLError("xml model defined non-positive time step")
+            raise ValueError("Invalid simulation timestep defined!")
         self.control_freq = control_freq
         if control_freq <= 0:
-            raise SimulationError(
-                "control frequency {} is invalid".format(control_freq)
-            )
+            raise SimulationError("Control frequency {} is invalid".format(control_freq))
         self.control_timestep = 1. / control_freq
 
     def set_model_postprocessor(self, postprocessor):
@@ -183,7 +187,7 @@ class MujocoEnv(metaclass=EnvMeta):
         if self._model_postprocessor is not None:
             self._model_postprocessor(self.model)
 
-    def _get_reference(self):
+    def _setup_references(self):
         """
         Sets up references to important components. A reference is typically an
         index or a list of indices that point to the corresponding elements
@@ -191,6 +195,15 @@ class MujocoEnv(metaclass=EnvMeta):
         """
         pass
 
+    def _setup_observables(self):
+        """
+        Sets up observables to be used for this environment.
+
+        Returns:
+            OrderedDict: Dictionary mapping observable names to its corresponding Observable object
+        """
+        return OrderedDict()
+
     def _initialize_sim(self, xml_string=None):
         """
         Creates a MjSim object and stores it in self.sim. If @xml_string is specified, the MjSim object will be created
@@ -229,10 +242,17 @@ class MujocoEnv(metaclass=EnvMeta):
         # Reset necessary robosuite-centric variables
         self._reset_internal()
         self.sim.forward()
+        # Setup observables, reloading if
+        self._obs_cache = {}
+        if self.hard_reset:
+            # If we're using hard reset, must re-update sensor object references
+            _observables = self._setup_observables()
+            for obs_name, obs in _observables.items():
+                self.modify_observable(observable_name=obs_name, attribute="sensor", modifier=obs._sensor)
         # Make sure that all sites are toggled OFF by default
         self.visualize(vis_settings={vis: False for vis in self._visualizations})
         # Return new observations
-        return self._get_observation()
+        return self._get_observations(force_update=True)
 
     def _reset_internal(self):
         """Resets simulation internal configurations."""
@@ -263,20 +283,68 @@ class MujocoEnv(metaclass=EnvMeta):
 
         # additional housekeeping
         self.sim_state_initial = self.sim.get_state()
-        self._get_reference()
+        self._setup_references()
         self.cur_time = 0
         self.timestep = 0
         self.done = False
 
-    def _get_observation(self):
+        # Empty observation cache and reset all observables
+        self._obs_cache = {}
+        for observable in self._observables.values():
+            observable.reset()
+
+    def _update_observables(self, force=False):
+        """
+        Updates all observables in this environment
+
+        Args:
+            force (bool): If True, will force all the observables to update their internal values to the newest
+                value. This is useful if, e.g., you want to grab observations when directly setting simulation states
+                without actually stepping the simulation.
+        """
+        for observable in self._observables.values():
+            observable.update(timestep=self.model_timestep, obs_cache=self._obs_cache, force=force)
+
+    def _get_observations(self, force_update=False):
         """
         Grabs observations from the environment.
 
+        Args:
+            force_update (bool): If True, will force all the observables to update their internal values to the newest
+                value. This is useful if, e.g., you want to grab observations when directly setting simulation states
+                without actually stepping the simulation.
+
         Returns:
             OrderedDict: OrderedDict containing observations [(name_string, np.array), ...]
 
         """
-        return OrderedDict()
+        observations = OrderedDict()
+        obs_by_modality = OrderedDict()
+
+        # Force an update if requested
+        if force_update:
+            self._update_observables(force=True)
+
+        # Loop through all observables and grab their current observation
+        for obs_name, observable in self._observables.items():
+            if observable.is_enabled() and observable.is_active():
+                obs = observable.obs
+                observations[obs_name] = obs
+                modality = observable.modality + "-state"
+                if modality not in obs_by_modality:
+                    obs_by_modality[modality] = []
+                # Make sure all observations are numpy arrays so we can concatenate them
+                array_obs = [obs] if type(obs) in {int, float} or not obs.shape else obs
+                obs_by_modality[modality].append(np.array(array_obs))
+
+        # Add in modality observations
+        for modality, obs in obs_by_modality.items():
+            # To save memory, we only concatenate the image observations if explicitly requested
+            if modality == "image-state" and not macros.CONCATENATE_IMAGES:
+                continue
+            observations[modality] = np.concatenate(obs, axis=-1)
+
+        return observations
 
     def step(self, action):
         """
@@ -314,13 +382,14 @@ class MujocoEnv(metaclass=EnvMeta):
             self.sim.forward()
             self._pre_action(action, policy_step)
             self.sim.step()
+            self._update_observables()
             policy_step = False
 
         # Note: this is done all at once to avoid floating point inaccuracies
         self.cur_time += self.control_timestep
 
         reward, done, info = self._post_action(action)
-        return self._get_observation(), reward, done, info
+        return self._get_observations(), reward, done, info
 
     def _pre_action(self, action, policy_step=False):
         """
@@ -384,7 +453,7 @@ class MujocoEnv(metaclass=EnvMeta):
         Returns:
             OrderedDict: Observations from the environment
         """
-        observation = self._get_observation()
+        observation = self._get_observations()
         return observation
 
     def clear_objects(self, object_names):
@@ -414,36 +483,6 @@ class MujocoEnv(metaclass=EnvMeta):
         for obj in self.model.mujoco_objects:
             obj.set_sites_visibility(sim=self.sim, visible=vis_settings["env"])
 
-    @property
-    def _visualizations(self):
-        """
-        Visualization keywords for this environment
-
-        Returns:
-            set: All components that can be individually visualized for this environment
-        """
-        return {"env"}
-
-    @property
-    def action_spec(self):
-        """
-        Action specification should be implemented in subclasses.
-
-        Action space is represented by a tuple of (low, high), which are two numpy
-        vectors that specify the min/max action limits per dimension.
-        """
-        raise NotImplementedError
-
-    @property
-    def action_dim(self):
-        """
-        Size of the action space
-
-        Returns:
-            int: Action space dimension
-        """
-        raise NotImplementedError
-
     def reset_from_xml_string(self, xml_string):
         """
         Reloads the environment from an XML description of the environment.
@@ -528,6 +567,55 @@ class MujocoEnv(metaclass=EnvMeta):
                 contact_set.add(g1)
         return contact_set
 
+    def add_observable(self, observable):
+        """
+        Adds an observable to this environment.
+
+        Args:
+            observable (Observable): Observable instance.
+        """
+        assert observable.name not in self._observables,\
+            "Observable name {} is already associated with an existing observable! Use modify_observable(...) " \
+            "to modify a pre-existing observable.".format(observable.name)
+        self._observables[observable.name] = observable
+
+    def modify_observable(self, observable_name, attribute, modifier):
+        """
+        Modifies observable with associated name @observable_name, replacing the given @attribute with @modifier.
+
+        Args:
+             observable_name (str): Observable to modify
+             attribute (str): Observable attribute to modify.
+                Options are {`'sensor'`, `'corrupter'`,`'filter'`,  `'delayer'`, `'sampling_rate'`,
+                `'enabled'`, `'active'`}
+             modifier (any): New function / value to replace with for observable. If a function, new signature should
+                match the function being replaced.
+        """
+        # Find the observable
+        assert observable_name in self._observables, "No valid observable with name {} found. Options are: {}".\
+            format(observable_name, self.observation_names)
+        obs = self._observables[observable_name]
+        # replace attribute accordingly
+        if attribute == "sensor":
+            obs.set_sensor(modifier)
+        elif attribute == "corrupter":
+            obs.set_corrupter(modifier)
+        elif attribute == "filter":
+            obs.set_filter(modifier)
+        elif attribute == "delayer":
+            obs.set_delayer(modifier)
+        elif attribute == "sampling_rate":
+            obs.set_sampling_rate(modifier)
+        elif attribute == "enabled":
+            obs.set_enabled(modifier)
+        elif attribute == "active":
+            obs.set_active(modifier)
+        else:
+            # Invalid attribute specified
+            raise ValueError("Invalid observable attribute specified. Requested: {}, valid options are {}".
+                             format(attribute, {"sensor", "corrupter", "filter", "delayer",
+                                                "sampling_rate", "enabled", "active"}))
+
     def _check_success(self):
         """
         Checks if the task has been completed. Should be implemented by subclasses
@@ -549,3 +637,76 @@ class MujocoEnv(metaclass=EnvMeta):
     def close(self):
         """Do any cleanup necessary here."""
         self._destroy_viewer()
+
+    @property
+    def observation_modalities(self):
+        """
+        Modalities for this environment's observations
+
+        Returns:
+            set: All observation modalities
+        """
+        return set([observable.modality for observable in self._observables.values()])
+
+    @property
+    def observation_names(self):
+        """
+        Grabs all names for this environment's observables
+
+        Returns:
+            set: All observation names
+        """
+        return set(self._observables.keys())
+
+    @property
+    def enabled_observables(self):
+        """
+        Grabs all names of enabled observables for this environment. An observable is considered enabled if its values
+        are being continually computed / updated at each simulation timestep.
+
+        Returns:
+            set: All enabled observation names
+        """
+        return set([name for name, observable in self._observables.items() if observable.is_enabled()])
+
+    @property
+    def active_observables(self):
+        """
+        Grabs all names of active observables for this environment. An observable is considered active if its value is
+        being returned in the observation dict from _get_observations() call or from the step() call (assuming this
+        observable is enabled).
+
+        Returns:
+            set: All active observation names
+        """
+        return set([name for name, observable in self._observables.items() if observable.is_active()])
+
+    @property
+    def _visualizations(self):
+        """
+        Visualization keywords for this environment
+
+        Returns:
+            set: All components that can be individually visualized for this environment
+        """
+        return {"env"}
+
+    @property
+    def action_spec(self):
+        """
+        Action specification should be implemented in subclasses.
+
+        Action space is represented by a tuple of (low, high), which are two numpy
+        vectors that specify the min/max action limits per dimension.
+        """
+        raise NotImplementedError
+
+    @property
+    def action_dim(self):
+        """
+        Size of the action space
+
+        Returns:
+            int: Action space dimension
+        """
+        raise NotImplementedError
